# frozen_string_literal: true

require 'websocket'
module Landline
  # Module that holds websocket primitives
  module WebSocket
    # Event system
    module Eventifier
      # Attach event listener
      # @param event [Symbol]
      # @param listener [#call]
      def on(event, &listener)
        @__listeners ||= {}
        @__listeners[event] ||= []
        @__listeners[event].append(listener)
        listener
      end

      # Attach event listener
      # @param event [Symbol]
      # @param listener [#call]
      def off(event, listener)
        @__listeners ||= {}
        @__listeners[event]&.delete(listener)
      end

      private

      # Trigger the queue clearing process
      # @return [void]
      def _process
        return if @processing

        @__processing = true
        @__queue ||= []
        @__listeners ||= {}
        until @__queue.empty?
          event, msg = @__queue.shift
          if @__listeners.include? event
            @__listeners[event].each { |x| x.call(*msg) }
          end
        end
        @processing = false
      end

      # Send internal event
      # @param event [Symbol]
      # @param data [Array]
      # @return [void]
      def _emit(event, *data)
        return unless @__listeners

        return unless @__listeners.include? event

        @__queue ||= []
        @__queue.push([event, data])
        _process
      end
    end

    # Socket-like object representing websocket interface
    class WSockWrapper
      include Eventifier

      class WebSocketError < StandardError
      end

      def initialize(io, version: 7)
        @io = io
        @version = version
        @frame_parser = ::WebSocket::Frame::Incoming::Server.new(
          version: version
        )
        @readable = true
        @writable = true
      end

      # Send data through websocket
      # @param data [String] binary data
      # @return [void]
      def write(data, type: :text)
        unless @writable
          raise self.class::WebSocketError,
                "socket closed for writing"
        end

        frame = ::WebSocket::Frame::Outgoing::Server.new(
          version: @version,
          data: data,
          type: type
        )
        @io.write(frame.to_s)
      rescue Errno::EPIPE => e
        @writable = false
        _emit :error, e
        close
        nil
      end

      # Read data from socket synchronously
      # @return [WebSocket::Frame::Base, nil] nil if socket received a close event
      def read
        unless @readable
          raise self.class::WebSocketError,
                "socket closed for reading"
        end

        _process_events(proc { _read })
      end

      # Read data from socket without blocking
      # @return [WebSocket::Frame::Base, nil] nil if socket received a close event
      def read_nonblock
        unless @readable
          raise self.class::WebSocketError,
                "socket closed for reading"
        end

        _process_events(proc { _read_nonblock })
      end

      # Establish a connection through handshake
      # @return [self]
      def self.handshake(request, version: 7, **opts)
        raise StandardError, "stream cannot be hijacked" unless request.hijack

        handshake = create_handshake(request, version: version, **opts)
        return nil unless handshake

        io = request.hijack.call
        io.sendmsg(handshake.to_s)
        new(io, version: version)
      end

      # Initiate a handshake
      def self.create_handshake(request, **opts)
        handshake = ::WebSocket::Handshake::Server.new(**opts)
        handshake.from_hash({
                              headers: request.headers,
                              path: request.path_info,
                              query: request.query.query,
                              body: request.body
                            })
        return nil unless handshake.finished? and handshake.valid?

        handshake
      end

      # Close socket for reading
      def close_read
        raise WebSocketError, 'socket closed for reading' unless @readable

        _emit :close
        @readable = false
        @io.close_read
      end

      # Close socket for reading
      def close_write
        raise WebSocketError, 'socket closed for writing' unless @writable

        write(nil, type: :close)
        @writable = false
        @io.close_write
      end

      # Close the socket entirely
      def close
        raise WebSocketError, 'socket closed' unless @writable or @readable

        close_read if @readable
        close_write if @writable
      end

      # Obtain internal IO object
      # @return [IO]
      def to_io
        io
      end

      attr_reader :io, :readable, :writable

      private

      # Process incoming websocket events
      # @param next_frame [#call] callback to get the next frame
      # @return [WebSocket::Frame::Base, nil]
      def _process_events(next_frame)
        loop do
          frame = next_frame.call
          case frame.type
          when :binary, :text, :pong then return frame
          when :ping
            write frame.to_s, type: :pong
          when :close
            close_read
            return nil
          else raise WebSocketError, "unknown frame type #{frame.type}"
          end
        end
      end

      # Receive data through websocket
      # @return [String] output from frame
      def _read
        while (char = @io.read(1))
          @frame_parser << char
          frame = @frame_parser.next
          return frame if frame
        end
      rescue Errno::ECONNRESET => e
        _emit :error, e
        close
        nil
      end

      # Receive data through websocket asynchronously
      # @return [String] output from frame
      def _read_nonblock
        while (char = @io.read_nonblock(1))
          @frame_parser << char
          frame = @frame_parser.next
          return frame if frame
        end
      rescue Errno::ECONNRESET => e
        _emit :error, e
        close
        nil
      end
    end
  end
end

module Landline
  module Handlers
    # Web socket server handler
    class WebSockServer < Landline::Probe
      # @param path [Object]
      # @param parent [Landline::Node]
      # @param params [Hash] options hash
      # @param callback [#call] callback to process request within
      # @option params [Integer] :version protocol version
      # @option params [Array<String>] :protocols array of supported sub-protocols
      # @option params [Boolean] :secure true if the server will use wss:// protocol
      def initialize(path, parent:, **params, &callback)
        nodeparams = params.dup
        nodeparams.delete(:version)
        nodeparams.delete(:protocols)
        nodeparams.delete(:secure)
        super(path, parent: parent, **nodeparams)
        @callback = callback
        @params = params
      end

      # Method callback on successful request navigation
      # Creates a websocket from a given request
      # @param request [Landline::Request]
      def process(request)
        @callback.call(
          Landline::WebSocket::WSockWrapper.handshake(
            request,
            **@params
          )
        )
        ""
      end
    end
  end

  module DSL
    module PathConstructors
      # (in Landline::Path context)
      # Create a new websocket handler
      def websocket(path, **args, &setup)
        register(Landline::Handlers::WebSockServer.new(path,
                                                       parent: @origin,
                                                       **args,
                                                       &setup))
      end
    end
  end
end
