# frozen_string_literal: true

require_relative 'pattern_matching'
require_relative 'node'
require_relative 'dsl/constructors_path'
require_relative 'dsl/methods_path'
require_relative 'util/lookup'

module Landline
  # Execution context for path setup block.
  class PathContext
    include Landline::DSL::PathConstructors
    include Landline::DSL::PathMethods

    def initialize(path)
      @origin = path
    end
  end

  # Primary building block of request navigation.
  class Path < Landline::Node
    Context = Landline::PathContext

    # @param path [Object] Object to generate {Landline::Pattern} from
    # @param parent [Landline::Node] Parent object to inherit properties to
    # @param setup [#call] Setup block
    def initialize(path, parent:, **params, &setup)
      super(path, parent: parent, **params)
      # Child nodes array
      @children = []
      # Arrays of preprocessors, postprocessors and filters
      @preprocessors = []
      @postprocessors = []
      @filters = []
      # Contexts setup
      context = self.class::Context.new(self)
      context.instance_exec(&setup)
    end

    # (see ::Landline::Node#go)
    def go(request)
      # This is done to allow pipeline to interject handlers
      # I'm more than willing to admit that this is stupid,
      # but it is well worth the logical flexibility.
      if ['handle.default', 'handle.505'].any? do |x|
           @properties.storage.include? x
         end
        begin
          super(request)
        rescue StandardError => e
          _die(request, 500, backtrace: [e.to_s] + e.backtrace, error: e)
        end
      else
        super(request)
      end
    end

    # Method callback on successful request navigation.
    # Finds the next appropriate path to go to.
    # @return [Boolean] true if further navigation will be done
    # @raise [UncaughtThrowError] by default throws :response if no matches found.
    def process(request)
      if @pipeline
        @pipeline.call(request) { |inner_req| process_wrapped(inner_req) }
      else
        process_wrapped(request)
      end
    end

    # Add a preprocessor to the path.
    # Does not modify path execution.
    # @param block [#call]
    # @yieldparam request [Landline::Request]
    def preprocess(&block)
      @preprocessors.append(block)
    end

    # Add a postprocessor to the path.
    # @param block [#call]
    # @yieldparam request [Landline::Request]
    # @yieldparam response [Landline::Response]
    def postprocess(&block)
      @postprocessors.append(block)
    end

    # Add a filter to the path.
    # Blocks path access if a filter returns false.
    # @param block [#call]
    # @yieldparam request [Landline::Request]
    def filter(&block)
      @filters.append(block)
    end

    attr_reader :children, :properties

    attr_accessor :bounce, :pipeline

    private

    # Create an execution context for in-path processing blocks
    def get_context(request)
      request.context.origin.properties.lookup = @properties
      request.context
    end

    # Sequentially run through all filters and drop request if one is false
    # @param request [Landline::Request]
    # @return [Boolean] true if request passed all filters
    def run_filters(request)
      proccontext = get_context(request)
      @filters.each do |filter|
        output = catch(:break) do
          proccontext.instance_exec(request, &filter)
        end
        return false unless output
      end
      true
    end

    # Sequentially run all preprocessors on a request
    # @param request [Landline::Request]
    def run_preprocessors(request)
      proccontext = get_context(request)
      @preprocessors.each do |preproc|
        output = catch(:break) do
          proccontext.instance_exec(request, &preproc)
          true
        end
        return false unless output
      end
      true
    end

    # Append postprocessors to request
    # @param request [Landline::Request]
    def enqueue_postprocessors(request)
      request.postprocessors.append(*@postprocessors)
    end

    # Method callback on successful request navigation.
    # Finds the next appropriate path to go to.
    # (inner pipeline-wrapped handler)
    # @return [Boolean] true if further navigation will be done
    # @raise [UncaughtThrowError] by default throws :response if no matches found.
    def process_wrapped(request)
      return false unless run_filters(request)

      return false unless run_preprocessors(request)

      enqueue_postprocessors(request)
      @children.each do |x|
        value = x.go(request)
        return exit_stack(request, value) if value
      end
      value = index(request)
      return exit_stack(request, value) if value

      notfound(request)
    end

    # Run enqueued postprocessors on navigation failure
    # @param request [Landline::Request]
    def exit_stack(request, response = nil)
      request.run_postprocessors(response)
      response
    end

    # Exit with failure or throw 404
    def notfound(request)
      @bounce ? exit_stack(request) : _die(request, 404)
    end

    # Try to perform indexing on the path if possible
    # @param request [Landline::Request]
    # @return [Boolean] true if indexing succeeded
    def index(request)
      return false unless request.path.match?(/^\/?$/)

      @properties["index"].each do |index|
        request.path = index
        @children.each do |x|
          if (value = x.go(request))
            return value
          end
        end
      end
      false
    end

    # Handle an errorcode
    # @param errorcode [Integer]
    # @param backtrace [Array(String), nil]
    # @raise [UncaughtThrowError] throws :finish to stop processing
    def _die(request, errorcode, backtrace: nil, error: nil)
      proccontext = get_context(request)
      response = Landline::Response.convert(
        proccontext.instance_exec(
          errorcode,
          backtrace: backtrace,
          error: error,
          &(@properties["handle.#{errorcode}"] or
            @properties["handle.default"])
        )
      )
      response.status = errorcode if response.status == 200
      throw :finish, response
    end
  end
end
